package net.xdocc;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

public class Cache
{
	private final Map<File, CacheItem> cache;
	private final File cacheFile;
	private volatile boolean changed = false;
	private final ScheduledExecutorService executorService = Executors
			.newSingleThreadScheduledExecutor();
	private final File generatedPath;

	public Cache(File generatedPath)
	{
		this.generatedPath = generatedPath;
		this.cacheFile = new File(generatedPath, ".cache");
		this.cache = createOrGetCache(cacheFile);
		executorService.scheduleWithFixedDelay(new Runnable()
		{
			@Override
			public void run()
			{
				if (changed)
				{
					try
					{
						saveCache0();
						changed = false;
					}
					catch (IOException e)
					{
						e.printStackTrace();
					}
				}
			}
		}, 2, 2, TimeUnit.SECONDS);
	}

	public void shutdown()
	{
		executorService.shutdown();
	}

	public void addGenerated(File source, File generated) throws IOException
	{
		synchronized (this)
		{
			//System.err.println("add generated " + generated + " for " + source);
			CacheItem cacheItem = cache.get(source);
			if (cacheItem == null)
			{
				cacheItem = new CacheItem();
				cache.put(source, cacheItem);
			}
			cacheItem.setSourceLastModified(source);
			cacheItem.addGenerated(generated);
			saveCache();
		}
	}

	public List<File> getGenerated(File backingFile)
	{
		List<File> result = new ArrayList<File>();
		synchronized (this)
		{
			CacheItem cacheItem = cache.get(backingFile);
			if (cacheItem != null)
			{
				for (File file : cacheItem.getGenerated())
					result.add(file);
			}
		}
		return result;
	}

	public void removeIndexGenerated(File source) throws IOException
	{
		synchronized (this)
		{
			CacheItem cacheItem = cache.get(source);
			if (cacheItem == null)
				return;
			for (File generated : cacheItem.getGenerated())
			{
				if (generated.getName().startsWith("index.dir.html"))
				{
					generated.delete();
					System.err.println("removed generated file " + generated);
					saveCache();
				}
			}
		}
	}

	// static int i=0;
	public void removeSource(File source) throws IOException
	{
		synchronized (this)
		{
			CacheItem cacheItem = cache.get(source);
			if (cacheItem == null)
				return;
			// if(source.lastModified()==cacheItem.getSourceLastModified())
			// return;
			cache.remove(source);
			for (File generated : new ArrayList<File>(cacheItem.getGenerated()))
			{
				generated.delete();
				System.err.println("removed generated2 file " + generated + " from source "
						+ source);
			}
			saveCache();
			// i++;
			// if(i>3)
			// {
			// for(StackTraceElement s:Thread.currentThread().getStackTrace())
			// System.err.println(s.toString());
			// }
		}
	}

	public List<File> getSources(File generated)
	{
		List<File> result = new ArrayList<File>();
		for (Map.Entry<File, CacheItem> entry : cache.entrySet())
		{
			if (entry.getValue().contains(generated))
				result.add(entry.getKey());
		}
		return result;
	}

	public boolean isValid(File source)
	{
		return isValid(source, new ArrayList<File>());
	}

	public boolean isValid(File source, File shouldBeGenerated)
	{
		List<File> tmp = new ArrayList<File>();
		tmp.add(shouldBeGenerated);
		return isValid(source, tmp);
	}

	public boolean isValid(File source, List<File> shouldBeGenerated)
	{
		synchronized (this)
		{
			CacheItem cacheItem = cache.get(source);
			if (cacheItem == null)
				return false;
			if (cacheItem.getSourceLastModified() != source.lastModified())
				return false;
			boolean retVal = true;
			for (Iterator<File> iterator = cacheItem.getGenerated().iterator(); iterator.hasNext();)
			{
				File generated = iterator.next();
				shouldBeGenerated.remove(generated);
				if (!generated.exists())
				{
					iterator.remove();
					retVal = false;
				}
				// long mod = genFile.lastModified();
				// if (mod != entry.getValue().longValue())
				// return false;
			}
			if (shouldBeGenerated.size() > 0)
				return false;
			for (Iterator<Map.Entry<File, Long>> iterator2 = cacheItem.getDependencies().entrySet()
					.iterator(); iterator2.hasNext();)
			{
				Map.Entry<File, Long> entry = iterator2.next();
				if (!entry.getKey().exists())
				{
					iterator2.remove();
					retVal = false;
				}
				// check dependecies, if they changed
				else if (entry.getValue() != entry.getKey().lastModified())
					retVal = false;
			}
			return retVal;
		}
	}

	private void saveCache() throws IOException
	{
		changed = true;
	}

	private void saveCache0() throws IOException
	{
		synchronized (this)
		{
			FileOutputStream fos = new FileOutputStream(cacheFile);
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(cache);
			oos.close();
			fos.close();
		}
	}

	@SuppressWarnings("unchecked")
	private Map<File, CacheItem> createOrGetCache(File generatedFile)
	{
		if (cacheFile.exists())
		{
			try
			{
				FileInputStream fis = new FileInputStream(cacheFile);
				ObjectInputStream ois = new ObjectInputStream(fis);
				Object o = ois.readObject();
				Map<File, CacheItem> cache = (Map<File, CacheItem>) o;
				ois.close();
				fis.close();
				// remove every generated file if the source is not present
				System.err.println("loaded cache ");
				Set<File> sourceValidGenerated = removeUnusedGenerated(cache);
				removeUnusedGenerated(sourceValidGenerated);
				return cache;
			}
			catch (Exception e)
			{
				e.printStackTrace();
				return new HashMap<File, CacheItem>();
			}
		}
		else
		{
			System.err.println("no cached file");
			return new HashMap<File, CacheItem>();
		}
	}

	private Set<File> removeUnusedGenerated(Map<File, CacheItem> cache) throws IOException
	{
		Set<File> generated = new HashSet<File>();
		for (Iterator<Map.Entry<File, CacheItem>> iterator = cache.entrySet().iterator(); iterator
				.hasNext();)
		{
			Map.Entry<File, CacheItem> entry = iterator.next();
			File source = entry.getKey();
			System.err.println("check " + source);
			CacheItem cacheItem = entry.getValue();
			if (!source.exists())
			{
				for (File tmp : cacheItem.getGenerated())
				{
					tmp.delete();
					System.err.println("removed generated file " + tmp);
				}
				iterator.remove();
				saveCache();
			}
			else
			{
				generated.addAll(cacheItem.getGenerated());
			}
		}
		return generated;
	}

	/**
	 * Removes any additional files in the generated directory, which should not
	 * be there.
	 * 
	 * @param sourceValidGenerated The set of files that should be there
	 */
	private void removeUnusedGenerated(Set<File> sourceValidGenerated)
	{
		removeUnusedGeneratedRec(sourceValidGenerated, generatedPath);
	}

	private void removeUnusedGeneratedRec(final Set<File> sourceValidGenerated, File path)
	{
		path.listFiles(new FileFilter()
		{
			@Override
			public boolean accept(File pathname)
			{
				if (pathname.isDirectory())
					removeUnusedGeneratedRec(sourceValidGenerated, pathname);
				else
				{
					if (!sourceValidGenerated.contains(pathname) && !pathname.equals(cacheFile))
						pathname.delete();
				}
				return false;
			}
		});
	}

	public int generatedCount(File backingFile)
	{
		synchronized (this)
		{
			CacheItem cacheItem = cache.get(backingFile);
			if (cacheItem != null)
				return cacheItem.generatedCount();
		}
		return 0;
	}

	public void addDependency(File source, File dependecy)
	{
		synchronized (this)
		{
			CacheItem cacheItem = cache.get(source);
			if (cacheItem != null)
				cacheItem.addDependency(dependecy);
		}
	}
}
