package net.xdocc;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Timer;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.LogManager;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.xdocc.virtualfilesystem.VirtualDirectory;
import net.xdocc.virtualfilesystem.VirtualFS;
import net.xdocc.virtualfilesystem.VirtualFile;
import net.xdocc.virtualfilesystem.VirtualFileListener;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.io.FileUtils;
import org.apache.velocity.app.VelocityEngine;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import eu.medsea.mimeutil.MimeUtil;

public class Compiler
{
	final private static Logger logger = LoggerFactory.getLogger(Compiler.class);
	private final static Options options = new Options();
	// final private static Properties properties = new Properties();
	final private static String SITE_SKELETON_ROOT = "default-site";
	final private static String[] SITE_SKELETON_FILES = new String[] { "content/design/layout.css",
			"settings/" + Site.SETTINGS_NAME, "settings/" + Site.TEMPLATE_NAME };
	final private Timer timer = new Timer(true);
	private final ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime()
			.availableProcessors() + 1);
	private final ScheduledExecutorService executorService = Executors
			.newSingleThreadScheduledExecutor();
	final private Map<String, Site> sites = new HashMap<String, Site>();
	private static PropertyReader propertyReader;
	static
	{
		options.addOption("c", "config", true, "set configuration file");
		MimeUtil.registerMimeDetector("eu.medsea.mimeutil.detector.ExtensionMimeDetector");
		try
		{
			InputStream is=Compiler.class.getResourceAsStream("log.properties");
			if(is==null)
				is=new FileInputStream("/etc/xdocc/log.properties");
			LogManager.getLogManager().readConfiguration(is);
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
	}
	
	public static void dummy()
	{
		System.err.println("dummy");
	}

	public static void main(String[] args) throws Exception
	{
		CommandLineParser parser = new PosixParser();
		CommandLine cmd = parser.parse(options, args);
		String file = cmd.getOptionValue("c");
		if (file == null)
		{
			HelpFormatter hf = new HelpFormatter();
			hf.printHelp("java", options);
			return;
		}
		loadConfiguration(new File(file));
		new Compiler(propertyReader);
	}

	private static void loadConfiguration(File config) throws IOException
	{
		if (!config.exists())
		{
			logger.error("xdocc.properties not found");
			throw new IOException("xdocc.properties not found");
		}
		propertyReader = new PropertyReader(config);
	}

	public Compiler(PropertyReader propertyReader) throws Exception
	{
		List<Properties> props = propertyReader.getTag("site");
		for (Properties p : props)
		{
			String key = PropertyReader.getKey(p);
			File siteRoot = new File(p.getProperty("source"));
			File generated = new File(p.getProperty("generated"));
			final Site site = createSite(siteRoot, generated);
			addTemplateListener(site);
			sites.put(key, site);
			VirtualFileListener vfs = createCompileListener(site);
			site.getTemplateVirtualFS().addVirtualFileListener(vfs);
			site.getContentVirtualFS().addVirtualFileListener(vfs);
			compileDir(site, site.getContentVirtualFS().getRoot());
			executorService.scheduleWithFixedDelay(new Runnable()
			{
				@Override
				public void run()
				{
					if (site.isChanged())
					{
						try
						{
							List<Future<Boolean>> tmp = compileDir(site, site.getContentVirtualFS()
									.getRoot());
							for (Future<Boolean> future : tmp)
								future.get();
							site.setChanged(false);
						}
						catch (Exception e)
						{
							logger.error("compiler error: " + e);
							e.printStackTrace();
						}
					}
				}
			}, 3, 3, TimeUnit.SECONDS);
		}
	}

	private List<Future<Boolean>> compileDir(Site site, VirtualDirectory directoryToCompile)
			throws Exception
	{
		List<Future<Boolean>> result = new ArrayList<Future<Boolean>>();
		System.err.println("compile " + directoryToCompile.getRelativeVisibleFile());
		if (directoryToCompile.getRelativeVisibleFile().equals("/hallo/weblog/"))
			System.err.println("compile1 ");
		// TODO:make proper protection
		if (!checkProtect(site, directoryToCompile))
		{
			// only compile a directory if it is visible or if its root, the
			// compile will generate an index.dir.html file
			if (directoryToCompile.getAttributes().isVisible() || directoryToCompile.isRoot())
			{
				DirectoryCompiler dc = new DirectoryCompiler(site, directoryToCompile,
						directoryToCompile);
				result.add(executor.submit(dc));
			}
			for (VirtualFile vfile : directoryToCompile.getAccessibleFiles())
			{
				if (vfile.getAttributes().isVisible() || directoryToCompile.isAllVisible())
				{
					ContentCompiler cc = Contents.getFileHandler(site, vfile);
					result.add(executor.submit(cc));
				}
				else
				{
					// if not visible, copy only
					ContentCopy cc = new ContentCopy(site, vfile);
					result.add(executor.submit(cc));
				}
			}
		}
		// recursion
		for (VirtualDirectory vdir : directoryToCompile.getAccessibleDirectories())
			compileDir(site, vdir);
		return result;
	}

	private Site createSite(File sourceFile, File generated) throws Exception
	{
		File generatedFile = generated;
		File contentFile = new File(sourceFile, "content");
		File settingsFile = new File(sourceFile, "settings");
		//
		boolean atLestOne = true;
		atLestOne &= generatedFile.mkdirs();
		atLestOne &= sourceFile.mkdirs();
		atLestOne &= contentFile.mkdirs();
		atLestOne &= settingsFile.mkdirs();
		if (atLestOne)
		{
			for (String name : SITE_SKELETON_FILES)
			{
				URL url = Compiler.class.getResource(SITE_SKELETON_ROOT + "/" + name);
				File destination = new File(sourceFile, name);
				FileUtils.copyURLToFile(url, destination);
			}
		}
		Cache cacheNode = new Cache(generatedFile);
		VirtualFS templateVirtualFS = new VirtualFS(timer, settingsFile, 1000);
		VirtualFS contentVirtualFS = new VirtualFS(timer, contentFile, 1000);
		final VelocityEngine ve = createVelocityTemplateEngine(settingsFile);
		return new Site(contentVirtualFS, templateVirtualFS, generatedFile, ve, new File(
				settingsFile, Site.SETTINGS_NAME), cacheNode);
	}

	private VelocityEngine createVelocityTemplateEngine(File templateFile) throws Exception
	{
		VelocityEngine ve = new VelocityEngine();
		ve.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM, new Slf4jLogChute());
		ve.setProperty("file.resource.loader.path", templateFile.toString());
		ve.setProperty("file.resource.loader.cache", false);
		ve.init();
		return ve;
	}

	public void shutdown()
	{
		executorService.shutdown();
		executor.shutdown();
		for (Site site : sites.values())
		{
			site.shutdown();
		}
	}

	// TODO: do this properly
	private boolean checkProtect(final Site site, VirtualDirectory virtualDirectory)
			throws UnsupportedEncodingException
	{
		return findDirPattern(site, virtualDirectory, "htdigest_pass");
	}

	private VirtualFileListener createCompileListener(final Site site)
	{
		return new VirtualFileListener()
		{
			@Override
			public void virtualFileRemoved(VirtualFile virtualFile) throws Exception
			{
				site.getCache().removeSource(virtualFile.getBackingFile());
				site.setChanged(true);
			}

			@Override
			public void virtualFileChanged(VirtualFile virtualFile) throws Exception
			{
				site.getCache().removeSource(virtualFile.getBackingFile());
				site.setChanged(true);
			}

			@Override
			public void virtualFileAdded(VirtualFile virtualFile) throws Exception
			{
				site.setChanged(true);
			}

			@Override
			public void virtualDirectoryRemoved(VirtualDirectory virtualDirectory) throws Exception
			{
				site.getCache().removeSource(virtualDirectory.getBackingFile());
				site.setChanged(true);
			}

			@Override
			public void virtualDirectoryAdded(VirtualDirectory virtualDirectory) throws Exception
			{
				// notifyChanged(site, virtualDirectory);
				site.setChanged(true);
				addProperties(site, virtualDirectory);
			}

			@Override
			public void virtualDirectoryChanged(VirtualDirectory virtualDirectory) throws Exception
			{
				site.getCache().removeSource(virtualDirectory.getBackingFile());
				site.setChanged(true);
			}

			@Override
			public void virtualFileInDirectoryChanged(VirtualDirectory virtualDirectory)
					throws Exception
			{
				// remove the index in this directory
				site.getCache().removeIndexGenerated(virtualDirectory.getBackingFile());
				site.setChanged(true);
			}

			@Override
			public void exceptionCaught(Throwable t)
			{
				logger.error("Error " + t);
				t.printStackTrace();
			}
		};
	}

	private boolean findDirPattern(Site site, VirtualDirectory virtualDirectory, String name)
			throws UnsupportedEncodingException
	{
		List<Properties> p = site.getProperties().getTag("dir");
		if (p != null)
		{
			for (Properties properties : p)
			{
				if (properties.containsKey(name))
				{
					String pat = PropertyReader.getKey(properties);
					Pattern pattern = Pattern.compile(pat.trim());
					if (virtualDirectory.isRoot())
						return false;
					Matcher m = pattern.matcher(virtualDirectory.getRelativeVisibleFile());
					if (m.matches())
					{
						virtualDirectory.setProtected(true);
						return true;
					}
				}
			}
		}
		return false;
	}

	protected void addProperties(Site site, VirtualDirectory virtualDirectory)
			throws UnsupportedEncodingException
	{
		if (findDirPattern(site, virtualDirectory, "accessible_files_visible"))
			virtualDirectory.setAllVisible(true);
	}

	public Site getSite(String key)
	{
		return sites.get(key);
	}

	private void addTemplateListener(final Site site)
	{
		site.getTemplateVirtualFS().addVirtualFileListener(new VirtualFileListener()
		{
			@Override
			public void virtualFileRemoved(VirtualFile virtualFile) throws Exception
			{
			}

			@Override
			public void virtualFileInDirectoryChanged(VirtualDirectory virtualDirectory)
					throws Exception
			{
			}

			@Override
			public void virtualFileChanged(VirtualFile virtualFile) throws Exception
			{
				if (virtualFile.getAttributes().getName().startsWith("template"))
					site.readTemplate();
			}

			@Override
			public void virtualFileAdded(VirtualFile virtualFile) throws Exception
			{
			}

			@Override
			public void virtualDirectoryRemoved(VirtualDirectory virtualDirectory) throws Exception
			{
			}

			@Override
			public void virtualDirectoryChanged(VirtualDirectory virtualDirectory) throws Exception
			{
				// TODO Auto-generated method stub
			}

			@Override
			public void virtualDirectoryAdded(VirtualDirectory virtualDirectory) throws Exception
			{
				// TODO Auto-generated method stub
			}

			@Override
			public void exceptionCaught(Throwable t)
			{
				// TODO Auto-generated method stub
			}
		});
	}
}
