package net.xdocc;
import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.List;

import net.xdocc.virtualfilesystem.VirtualDirectory;
import net.xdocc.virtualfilesystem.VirtualFile;
import net.xdocc.virtualfilesystem.VirtualNode;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.eclipse.mylyn.internal.wikitext.textile.core.Textile;
import org.eclipse.mylyn.internal.wikitext.textile.core.TextileContentState;
import org.eclipse.mylyn.internal.wikitext.textile.core.phrase.EscapeTextilePhraseModifier;
import org.eclipse.mylyn.internal.wikitext.textile.core.phrase.ImageTextilePhraseModifier;
import org.eclipse.mylyn.internal.wikitext.textile.core.phrase.SimpleTextilePhraseModifier;
import org.eclipse.mylyn.internal.wikitext.textile.core.phrase.SimpleTextilePhraseModifier.Mode;
import org.eclipse.mylyn.wikitext.core.parser.Attributes;
import org.eclipse.mylyn.wikitext.core.parser.ImageAttributes;
import org.eclipse.mylyn.wikitext.core.parser.LinkAttributes;
import org.eclipse.mylyn.wikitext.core.parser.MarkupParser;
import org.eclipse.mylyn.wikitext.core.parser.DocumentBuilder.SpanType;
import org.eclipse.mylyn.wikitext.core.parser.builder.HtmlDocumentBuilder;
import org.eclipse.mylyn.wikitext.core.parser.markup.PatternBasedElement;
import org.eclipse.mylyn.wikitext.core.parser.markup.PatternBasedElementProcessor;
import org.eclipse.mylyn.wikitext.core.parser.markup.phrase.HtmlEndTagPhraseModifier;
import org.eclipse.mylyn.wikitext.core.parser.markup.phrase.HtmlStartTagPhraseModifier;
import org.eclipse.mylyn.wikitext.textile.core.TextileLanguage;

public class TextileHandler implements ContentCompiler
{
	final private Site site;
	final private VirtualFile virtualNode;

	public TextileHandler(Site site, VirtualFile virtualNode)
	{
		this.site = site;
		this.virtualNode = virtualNode;
	}

	@Override
	public Boolean call() throws Exception
	{
		if (!virtualNode.getBackingFile().exists())
		{
			System.err.println("the file " + virtualNode + " does not exist!!");
			return false;
		}
		// do not compile files that do not have a date, these will
		// be compiled in index.dir.html
		if ((virtualNode.getParent().getVisibleFiles().size() <= 1 || !virtualNode.getParent()
				.hasFullEntries())
				&& (virtualNode.getParent().getAttributes().isVisible() || virtualNode.getParent()
						.isRoot()))
		{
			System.err.println("dont need to create " + virtualNode.getBackingFile());
			return false;
		}
		File generated = Utils.createGenerated(site, virtualNode, virtualNode.getAttributes()
				.getURLName()
				+ ".html");
		if (!site.getCache().isValid(virtualNode.getBackingFile(), generated))
		{
			Document doc = Utils.createDocument(site, virtualNode, virtualNode, virtualNode,0);
			VirtualFile virtualFile = (VirtualFile) virtualNode;
			Entry entry = view(site, virtualFile, virtualFile, virtualFile.getParent());
			doc.addContent(entry);
			Utils.writeHTML(site, virtualFile, virtualFile.getParent(), new Document[] { doc },
					generated);
		}
		return true;
	}

	@Override
	public Entry view(Site site, VirtualNode virtualNode, VirtualNode linkTo,
			VirtualDirectory current) throws Exception
	{
		if (!virtualNode.getBackingFile().exists())
		{
			System.err.println("the file " + virtualNode + " does not exist!!");
			return null;
		}
		String raw = FileUtils.readFileToString(virtualNode.getBackingFile());
		MarkupParser markupParser = new MarkupParser();
		markupParser.setMarkupLanguage(new MyTextileLanguage());
		StringWriter out = new StringWriter();
		String webRoot = Utils.createRelativePathWeb(site, current);
		MyHtmlDocumentBuilder htmlDocumentBuilder = new MyHtmlDocumentBuilder(out, site, webRoot,
				(VirtualFile) virtualNode);
		htmlDocumentBuilder.setEmitAsDocument(false);
		markupParser.setBuilder(htmlDocumentBuilder);
		markupParser.parse(raw);
		markupParser.setBuilder(null);
		return new Entry("<div class=\"textile_full\">" + out.toString() + "</div>", 1, virtualNode
				.getBackingFile());
	}

	@Override
	public List<Document> viewCompiled(Site site, VirtualNode virtualNode, VirtualNode linkTo,
			VirtualDirectory current)
	{
		return null;
	}
	private class MyTextileLanguage extends TextileLanguage
	{
		@Override
		protected void addStandardPhraseModifiers(PatternBasedSyntax phraseModifierSyntax)
		{
			boolean escapingHtml = configuration == null ? false : configuration
					.isEscapingHtmlAndXml();
			phraseModifierSyntax.add(new HtmlEndTagPhraseModifier(escapingHtml));
			phraseModifierSyntax.add(new HtmlStartTagPhraseModifier(escapingHtml));
			phraseModifierSyntax
					.beginGroup("(?:(?<=[\\s\\.,\\\"'?!;:\\)\\(\\{\\}\\[\\]])|^)(?:", 0); //$NON-NLS-1$
			phraseModifierSyntax.add(new EscapeTextilePhraseModifier());
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"**", SpanType.BOLD, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"??", SpanType.CITATION, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"__", SpanType.ITALIC, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"_", SpanType.EMPHASIS, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"*", SpanType.STRONG, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"+", SpanType.INSERTED, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"~", SpanType.SUBSCRIPT, Mode.NORMAL)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"^", SpanType.SUPERSCRIPT, Mode.NORMAL)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"@", SpanType.CODE, Mode.SPECIAL)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"%", SpanType.SPAN, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new SimpleTextilePhraseModifier(
					"-", SpanType.DELETED, Mode.NESTING)); //$NON-NLS-1$
			phraseModifierSyntax.add(new ImageTextilePhraseModifier());
			phraseModifierSyntax.add(new MyHyperlinkPhraseModifier()); // hyperlinks
			// are
			// actually
			// a
			// phrase
			// modifier
			// see
			// bug
			// 283093
			phraseModifierSyntax.endGroup(")(?=\\W|$)", 0); //$NON-NLS-1$
		}
	}
	private class MyHyperlinkPhraseModifier extends PatternBasedElement
	{
		protected static final int ATTRIBUTES_OFFSET = 2;

		@Override
		protected String getPattern(int groupOffset)
		{
			return "(?:(\")" + Textile.REGEX_ATTRIBUTES + "([^\"]+)\\" + (1 + groupOffset) + ":([^\\s]*[^\\s!.)(,:;]))"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}

		@Override
		protected int getPatternGroupCount()
		{
			return 3 + Textile.ATTRIBUTES_GROUP_COUNT;
		}

		@Override
		protected PatternBasedElementProcessor newProcessor()
		{
			return new HyperlinkProcessor();
		}
		private class HyperlinkProcessor extends PatternBasedElementProcessor
		{
			@Override
			public void emit()
			{
				String hyperlinkBoundaryText = group(1);
				String hyperlinkSrc = group(2 + Textile.ATTRIBUTES_GROUP_COUNT);
				String href = group(3 + Textile.ATTRIBUTES_GROUP_COUNT);
				if (href.startsWith("\"") && href.endsWith("\""))
					href = href.substring(1, href.length() - 1);
				String namedLinkUrl = ((TextileContentState) getState()).getNamedLinkUrl(href);
				if (namedLinkUrl != null)
				{
					href = namedLinkUrl;
				}
				if (hyperlinkBoundaryText.equals("\"")) { //$NON-NLS-1$
					LinkAttributes attributes = new LinkAttributes();
					attributes.setHref(href);
					Textile.configureAttributes(this, attributes, ATTRIBUTES_OFFSET, false);
					builder.beginSpan(SpanType.LINK, attributes);
					getMarkupLanguage().emitMarkupLine(parser, state, start(2), hyperlinkSrc, 0);
					builder.endSpan();
				}
				else
				{
					final ImageAttributes attributes = new ImageAttributes();
					Textile.configureAttributes(this, attributes, ATTRIBUTES_OFFSET, false);
					builder.imageLink(attributes, href, hyperlinkSrc);
				}
			}
		}
	}
	private class MyHtmlDocumentBuilder extends HtmlDocumentBuilder
	{
		private String htmlNsUri = "http://www.w3.org/1999/xhtml";
		private VirtualFile virtualFile;
		private String webRoot;
		Site site;

		public MyHtmlDocumentBuilder(Writer out, Site site, String webRoot, VirtualFile virtualFile)
		{
			super(out);
			this.virtualFile = virtualFile;
			this.webRoot = webRoot;
			this.site = site;
		}

		@Override
		public void image(Attributes attributes, String url)
		{
			VirtualFile image = virtualFile.getParent().findFile(url);
			if (image == null)
			{
				System.err.println("Image " + url + " not found!");
				return;
			}
			try
			{
				String[] names = ImageHandler.generate(site, image);
				writer.writeEmptyElement(htmlNsUri, "img");
				writer.writeAttribute("border", Integer.toString(0));
				writer.writeAttribute("src", webRoot + names[0]);
				if (attributes.getCssClass() != null)
					writer.writeAttribute("class", attributes.getCssClass());
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}

		// return new Entry("<div id=\"image\"><a href=\"" + origURL +
		// "\"><img src=\"" + thumbURL
		// + "\" border=\"0\"/></a><div id=\"caption\">"
		// +
		// FilenameUtils.getBaseName(virtualNode.getAttributes().getVisibleName())
		// + "</div></div>", 3);
		@Override
		public void imageLink(Attributes linkAttributes, Attributes imageAttributes, String href,
				String imageUrl)
		{
			VirtualFile image = virtualFile.getParent().findFile(imageUrl);
			try
			{
				String[] names = ImageHandler.generate(site, image);
				writer.writeStartElement(htmlNsUri, "div");
				writer.writeAttribute("id", "image");
				if (imageAttributes.getCssClass() != null)
					writer.writeAttribute("class", imageAttributes.getCssClass());
				writer.writeStartElement(htmlNsUri, "a");
				emitAnchorHref(webRoot + names[2]);
				writer.writeEmptyElement(htmlNsUri, "img");
				writer.writeAttribute("border", Integer.toString(0));
				String src = webRoot + names[0];
				writer.writeAttribute("src", src);
				writer.writeEndElement(); // a
				if (image.getAttributes().isVisible())
				{
					writer.writeStartElement(htmlNsUri, "div");
					writer.writeAttribute("id", "caption");
					writer.writeCharacters(FilenameUtils.getBaseName(image.getAttributes()
							.getVisibleName()));
					writer.writeEndElement(); // div
				}
				writer.writeEndElement(); // div
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}

		@Override
		public void link(Attributes attributes, String hrefOrHashName, String text)
		{
			writer.writeStartElement(htmlNsUri, "a"); //$NON-NLS-1$
			emitAnchorHref(hrefOrHashName);
			// applyLinkAttributes(attributes, hrefOrHashName);
			characters(text);
			writer.writeEndElement(); // a
		}
	}
}
