package net.xdocc.virtualfilesystem;
import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class FileMonitor
{
	final private static Logger logger = LoggerFactory.getLogger(FileMonitor.class);
	final private Timer timer;
	final private Set<FileStat> files = new LinkedHashSet<FileStat>();
	final private Collection<FileListener> listeners = new ArrayList<FileListener>();
	final private long pollingIntervalMillis;

	public FileMonitor(long pollingIntervalMillis)
	{
		this(new Timer(true), pollingIntervalMillis);
	}

	public FileMonitor(Timer timer, long pollingIntervalMillis)
	{
		this.timer = timer;
		this.pollingIntervalMillis = pollingIntervalMillis;
	}

	public void init()
	{
		synchronized (timer)
		{
			if (logger.isDebugEnabled())
			{
				logger.debug("start monitoring for:");
				for (FileStat fs : files)
					logger.debug(fs.getFile().toString());
			}
			timer.schedule(new FileMonitorNotifier(), 0, pollingIntervalMillis);
		}
	}

	public Set<FileStat> stop()
	{
		synchronized (timer)
		{
			timer.cancel();
		}
		return files;
	}

	public void addFile(File file)
	{
		addFile(new FileStat(file, true));
	}

	private void addFile(FileStat fileStat)
	{
		synchronized (files)
		{
			files.add(fileStat);
		}
	}

	public void removeFile(File file)
	{
		removeFile(new FileStat(file));
	}

	private void removeFile(FileStat fileStat)
	{
		synchronized (files)
		{
			files.remove(fileStat);
		}
	}

	public void addListener(FileListener fileListener)
	{
		listeners.add(fileListener);
	}

	public void removeListener(FileListener fileListener)
	{
		listeners.remove(fileListener);
	}
	private class FileMonitorNotifier extends TimerTask
	{
		Collection<FileStat> filesAdded = new ArrayList<FileStat>();
		Collection<FileStat> filesChanged = new ArrayList<FileStat>();
		Collection<FileStat> filesRemoved = new ArrayList<FileStat>();
		//
		Collection<FileStat> checkAgain = new ArrayList<FileStat>();

		public void run()
		{
			boolean fileModified = false;
			try
			{
				synchronized (files)
				{
					for (FileStat fileStat : files)
					{
						long lastModifiedTime = fileStat.lastStoredModifiedTime();
						long newModifiedTime = fileStat.lastNewModified();
						//System.err.println("last="+lastModifiedTime+", new "+newModifiedTime);
						if (newModifiedTime == 0)
						{
							filesRemoved.add(fileStat);
							logger.debug("file removed " + fileStat.getFile());
						}
						else if (newModifiedTime != lastModifiedTime
								|| checkAgain.contains(fileStat))
						{
							if (fileStat.getFile().isFile())
							{
								fileModified = true;
								logger.debug("file changed " + fileStat.getFile());
								if(!filesChanged.contains(fileStat))
									filesChanged.add(fileStat);
								fileStat.updateLastModified();
							}
							// if files changed in the directory, its reported
							// by
							// notifychanged, if a file is deleted, its reported
							// by
							// notifyremoved. However, we need to take care of
							// new
							// files
							// now.
							else if (fileStat.getFile().isDirectory())
							{
								if (addFiles(filesAdded, fileStat))
								{
									checkAgain.add(fileStat);
									fileStat.updateLastModified();
								}
								else
									checkAgain.remove(fileStat);
							}
							
							// we need to check again, the next round
						}
					}
					//System.err.println("flag is "+fileModified);
				}
			}
			catch (Exception e)
			{
				notifyException(e);
				e.printStackTrace();
				logger.error("Failure in file iterator " + e);
			}
			try
			{
				if (fileModified == false && filesChanged.size() > 0)
				{
					for (FileStat fileStat : filesChanged)
					{
						logger.debug("reporting file changed " + fileStat.getFile());
						notifyChanged(fileStat);
					}
					filesChanged.clear();
				}
				for (FileStat fileStat : filesRemoved)
				{
					logger.debug("reporting file removed " + fileStat.getFile());
					notifyRemoved(fileStat);
					removeFile(fileStat);
				}
				filesRemoved.clear();
				for (FileStat fileStat : filesAdded)
				{
					logger.debug("reporting file add " + fileStat.getFile());
					notifyAdded(fileStat);
					addFile(fileStat);
				}
				filesAdded.clear();
			}
			catch (Throwable e)
			{
				notifyException(e);
				e.printStackTrace();
				logger.equals(e);
			}
		}

		private boolean addFiles(Collection<FileStat> filesAdded, FileStat fileStat)
		{
			boolean retVal = false;
			for (File f : fileStat.getFile().listFiles())
			{
				if (f.getName().startsWith(".") || f.getName().endsWith("~"))
					continue;
				FileStat nf = new FileStat(f);
				// System.err.println("check " + f);
				if (!files.contains(nf) && !filesAdded.contains(nf))
				{
					logger.debug("add file " + f);
					filesAdded.add(nf);
					retVal = true;
					if (nf.getFile().isDirectory())
						addFiles(filesAdded, nf);
				}
			}
			return retVal;
		}

		private void notifyAdded(FileStat fileStat) throws Exception
		{
			for (FileListener fileListener : listeners)
				fileListener.fileAdded(fileStat);
		}

		private void notifyChanged(FileStat fileStat) throws Exception
		{
			for (FileListener fileListener : listeners)
				fileListener.fileChanged(fileStat);
		}

		private void notifyRemoved(FileStat fileStat) throws Exception
		{
			for (FileListener fileListener : listeners)
				fileListener.fileRemoved(fileStat);
		}

		private void notifyException(Throwable t)
		{
			for (FileListener fileListener : listeners)
				fileListener.exceptionCaught(t);
		}
	}
}
