package net.xdocc.virtualfilesystem;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

public class VirtualDirectory extends VirtualNode
{
	final private Set<VirtualFile> visibleChildFiles = new HashSet<VirtualFile>();
	final private TreeMap<String, VirtualFile> accesibleChildFiles = new TreeMap<String, VirtualFile>();
	final private Set<VirtualDirectory> visibleChildDirectories = new HashSet<VirtualDirectory>();
	final private Map<String, VirtualDirectory> accesibleChildDirectories = new HashMap<String, VirtualDirectory>();
	//
	final private TreeMap<String, VirtualFile> urlChildFiles = new TreeMap<String, VirtualFile>();
	final private TreeMap<String, VirtualDirectory> urlChildDirectories = new TreeMap<String, VirtualDirectory>();
	private boolean allVisible;
	private boolean protectedDir;
	private boolean orderAsc;

	public VirtualDirectory(File backingFile, VirtualDirectory parent, VirtualFS virtualFS)
			throws IOException
	{
		super(backingFile, parent, virtualFS);
		if (!backingFile.isDirectory())
			throw new IllegalArgumentException("File " + backingFile + " is not a file");
		if (parent != null)
			parent.addDirectory(this);
		if (getAttributes().getTags().contains("asc"))
			setOrderAsc(true);
	}

	public boolean isRoot()
	{
		return getParent() == null;
	}

	public void addFile(VirtualFile virtualFile) throws IOException
	{
		// ignore dot files
		if (virtualFile.isDotFile() || virtualFile.isTildeFile())
			return;
		if (urlChildFiles.containsKey(virtualFile.getAttributes().getURLName()))
		{
			VirtualFile conflict = urlChildFiles.get(virtualFile.getAttributes().getURLName());
			conflict.setConflict();
			virtualFile.setConflict();
			System.out.println("There is a conflict for "
					+ virtualFile.getAttributes().getURLName() + " for file "
					+ virtualFile.getAttributes().getVisibleName());
		}
		// setChanged(true);
		if (virtualFile.getAttributes().isVisible())
			visibleChildFiles.add(virtualFile);
		urlChildFiles.put(virtualFile.getAttributes().getURLName(), virtualFile);
		accesibleChildFiles.put(virtualFile.getAttributes().getName(), virtualFile);
	}

	public void addDirectory(VirtualDirectory dirToAdd) throws IOException
	{
		if (dirToAdd.isDotFile() || dirToAdd.isTildeFile())
			return;
		if (urlChildDirectories.containsKey(dirToAdd.getAttributes().getURLName()))
		{
			VirtualDirectory conflict = urlChildDirectories.get(dirToAdd.getAttributes()
					.getURLName());
			conflict.setConflict();
			dirToAdd.setConflict();
			System.out.println("There is a conflict for " + dirToAdd.getAttributes().getURLName()
					+ " for file " + dirToAdd.getAttributes().getVisibleName());
		}
		// setChanged(true);
		if (dirToAdd.getAttributes().isVisible())
			visibleChildDirectories.add(dirToAdd);
		urlChildDirectories.put(dirToAdd.getAttributes().getURLName(), dirToAdd);
		accesibleChildDirectories.put(dirToAdd.getAttributes().getName(), dirToAdd);
	}

	public void removeFile(VirtualFile virtualFile)
	{
		visibleChildFiles.remove(virtualFile.getAttributes().getVisibleName());
		urlChildFiles.remove(virtualFile.getAttributes().getURLName());
		accesibleChildFiles.remove(virtualFile.getAttributes().getName());
	}

	public void removeDirectory(VirtualDirectory vdir)
	{
		visibleChildDirectories.remove(vdir.getAttributes().getVisibleName());
		urlChildDirectories.remove(vdir.getAttributes().getURLName());
		accesibleChildDirectories.remove(vdir.getAttributes().getName());
	}

	public VirtualDirectory getAccessibleChildDirectory(String urlName)
	{
		return accesibleChildDirectories.get(urlName);
	}

	public VirtualFile getAccessibleChildFile(String urlName)
	{
		return accesibleChildFiles.get(urlName);
	}

	public List<VirtualDirectory> getVisibleDirectories()
	{
		List<VirtualDirectory> tmp = new ArrayList<VirtualDirectory>(visibleChildDirectories);
		sort(tmp, isOrderAsc());
		return tmp;
	}

	private void sort(List<? extends VirtualNode> tmp, boolean orderAsc2)
	{
		if (!orderAsc2)
			Collections.sort(tmp);
		else
		{
			Collections.sort(tmp, new Comparator<VirtualNode>()
			{
				@Override
				public int compare(VirtualNode o1, VirtualNode o2)
				{
					return o1.compareTo(o2) * -1;
				}
			});
		}
	}

	public Collection<VirtualDirectory> getAccessibleDirectories()
	{
		return new ArrayList<VirtualDirectory>(accesibleChildDirectories.values());
	}

	public List<VirtualFile> getVisibleFiles()
	{
		List<VirtualFile> tmp = new ArrayList<VirtualFile>(visibleChildFiles);
		sort(tmp, isOrderAsc());
		return tmp;
	}

	public Collection<VirtualFile> getAccessibleFiles()
	{
		return new ArrayList<VirtualFile>(accesibleChildFiles.values());
	}

	public String getRelativeVisiblePath()
	{
		StringBuilder sb = new StringBuilder();
		VirtualDirectory vdir = this;
		while (vdir.getParent() != null)
		{
			sb.insert(0, "/");
			sb.insert(0, vdir.getAttributes().getURLName());
			vdir = vdir.getParent();
		}
		return sb.toString();
	}

	public VirtualFile findFile(String url)
	{
		return virtualFS.findFile(url, this);
	}

	public void setProtected(boolean protectedDir)
	{
		this.protectedDir = protectedDir;
	}

	public boolean isProtected()
	{
		return protectedDir;
	}

	public boolean isAllVisible()
	{
		return allVisible;
	}

	public void setAllVisible(boolean allVisible)
	{
		this.allVisible = allVisible;
	}

	public void setOrderAsc(boolean orderAsc)
	{
		this.orderAsc = orderAsc;
	}

	public boolean isOrderAsc()
	{
		return orderAsc;
	}

	public boolean hasFullEntries()
	{
		boolean fullEntry = false;
		for (VirtualFile vfile : getVisibleFiles())
		{
			if (vfile.getAttributes().isDate())
			{
				fullEntry = true;
				break;
			}
		}
		return fullEntry;
	}
}
