package net.xdocc.virtualfilesystem;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Timer;

import net.xdocc.Utils;

public class VirtualFS implements FileListener
{
	final private File root;
	final private VirtualDirectory virtualRoot;
	final private FileMonitor fileMonitor;
	final private List<VirtualFileListener> listeners = new ArrayList<VirtualFileListener>();

	public VirtualFS(File pathToRoot) throws IOException
	{
		this(new Timer(true), pathToRoot, 1000);
	}

	public VirtualFS(File pathToRoot, int pollIntervallMillis) throws IOException
	{
		this(new Timer(true), pathToRoot, pollIntervallMillis);
	}

	public VirtualFS(Timer timer, File pathToRoot, int pollIntervallMillis) throws IOException
	{
		this.root = pathToRoot;
		this.fileMonitor = new FileMonitor(timer, pollIntervallMillis);
		this.virtualRoot = new VirtualDirectory(root, null, this);
		fileMonitor.addFile(root);
		fileMonitor.addListener(this);
		fileMonitor.init();
	}

	public void dispose()
	{
		fileMonitor.stop();
	}

	@Override
	public void fileAdded(FileStat fileStat) throws Exception
	{
		if (fileStat.isFile())
		{
			File parentFile = fileStat.getFile().getParentFile();
			VirtualDirectory vdir = findDirectory(parentFile);
			if (vdir == null)
				return;
			VirtualFile fileToAdd = new VirtualFile(fileStat.getFile(), vdir, this);
			notifyFileAdd(fileToAdd);
			notifyFileInDirectoryChanged(fileToAdd.getParent());
		}
		else if (fileStat.isDirectory())
		{
			File parentFile = fileStat.getFile().getParentFile();
			VirtualDirectory vdir = findDirectory(parentFile);
			if (vdir == null)
				return;
			VirtualDirectory dirToAdd = new VirtualDirectory(fileStat.getFile(), vdir, this);
			notifyDirectoryAdd(dirToAdd);
		}
	}

	@Override
	public void fileChanged(FileStat fileStat) throws Exception
	{
		if (fileStat.isFile())
		{
			VirtualFile vfile = findFile(fileStat.getFile());
			if (vfile == null)
				return;
			notifyFileChange(vfile);
			notifyFileInDirectoryChanged(vfile.getParent());
		}
		else if (fileStat.isDirectory())
		{
			VirtualDirectory vdir = findDirectory(fileStat.getFile());
			if (vdir == null)
				return;
			notifyDirectoryChanged(vdir);
		}
	}

	@Override
	public void fileRemoved(FileStat fileStat) throws Exception
	{
		if (fileStat.isFile())
		{
			VirtualFile vfile = findFile(fileStat.getFile());
			if (vfile == null)
				return;
			vfile.getParent().removeFile(vfile);
			notifyFileRemove(vfile);
			notifyFileInDirectoryChanged(vfile.getParent());
		}
		else if (fileStat.isDirectory())
		{
			VirtualDirectory vdir = findDirectory(fileStat.getFile());
			if (vdir == null)
				return;
			vdir.getParent().removeDirectory(vdir);
			notifyDirectoryRemove(vdir);
		}
	}

	@Override
	public void exceptionCaught(Throwable t)
	{
		t.printStackTrace();
	}

	public void addVirtualFileListener(VirtualFileListener virtualFileListener)
	{
		listeners.add(virtualFileListener);
	}

	public void removeVirtualFileListener(VirtualFileListener virtualFileListener)
	{
		listeners.remove(virtualFileListener);
	}
	
	private void notifyFileInDirectoryChanged(VirtualDirectory virtualDirectory)
	{
		if (virtualDirectory.isDotFile() || virtualDirectory.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualFileInDirectoryChanged(virtualDirectory);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
		
	}

	private void notifyDirectoryAdd(VirtualDirectory virtualDirectory)
	{
		if (virtualDirectory.isDotFile() || virtualDirectory.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualDirectoryAdded(virtualDirectory);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
	}

	private void notifyDirectoryRemove(VirtualDirectory virtualDirectory)
	{
		if (virtualDirectory.isDotFile() || virtualDirectory.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualDirectoryRemoved(virtualDirectory);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
	}

	private void notifyDirectoryChanged(VirtualDirectory virtualDirectory)
	{
		if (virtualDirectory.isDotFile() || virtualDirectory.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualDirectoryChanged(virtualDirectory);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
	}

	private void notifyFileAdd(VirtualFile virtualFile)
	{
		if (virtualFile.isDotFile() || virtualFile.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualFileAdded(virtualFile);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
	}

	private void notifyFileRemove(VirtualFile virtualFile)
	{
		if (virtualFile.isDotFile() || virtualFile.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualFileRemoved(virtualFile);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
	}

	private void notifyFileChange(VirtualFile virtualFile)
	{
		if (virtualFile.isDotFile() || virtualFile.isTildeFile())
			return;
		for (VirtualFileListener listener : listeners)
		{
			try
			{
				listener.virtualFileChanged(virtualFile);
			}
			catch (Exception e)
			{
				listener.exceptionCaught(e);
			}
		}
	}

	public VirtualDirectory getRoot()
	{
		return virtualRoot;
	}

	private VirtualFile findFile(File file)
	{
		String fullPath = file.getAbsolutePath();
		String relativePath = Utils.makeRelativePath(fullPath, root.getAbsolutePath());
		return findFile(relativePath);
	}

	private VirtualDirectory findDirectory(File file)
	{
		String fullPath = file.getAbsolutePath();
		String relativePath = Utils.makeRelativePath(fullPath, root.getAbsolutePath());
		return findDirectory(relativePath);
	}

	public VirtualFile findFile(String path)
	{
		return findFile(path, virtualRoot);
	}

	public VirtualFile findFile(String path, VirtualDirectory root)
	{
		StringTokenizer st = new StringTokenizer(path, "/");
		VirtualDirectory vdir = root;
		while (st.hasMoreTokens())
		{
			String token = st.nextToken();
			VirtualDirectory vdir2 = vdir.getAccessibleChildDirectory(token);
			if (vdir2 == null)
				return vdir.getAccessibleChildFile(token);
			else
				vdir = vdir2;
		}
		return null;
	}

	public VirtualDirectory findDirectory(String path)
	{
		StringTokenizer st = new StringTokenizer(path, "/");
		VirtualDirectory vdir = virtualRoot;
		while (st.hasMoreTokens())
		{
			String token = st.nextToken();
			VirtualDirectory vdir2 = vdir.getAccessibleChildDirectory(token);
			if (vdir2 == null)
				return null;
			else
				vdir = vdir2;
		}
		return vdir;
	}

	public VirtualFile waitFor(String name)
	{
		VirtualFile file = findFile(name);
		try
		{
			while (file == null)
			{
				synchronized (this)
				{
					wait(1000);
				}
			}
			return file;
		}
		catch (InterruptedException e)
		{
			e.printStackTrace();
			return null;
		}
		
	}
}
