package net.xdocc.virtualfilesystem;
import java.io.File;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Collection;
import java.util.StringTokenizer;
import java.util.zip.CRC32;

public class VirtualNode implements Comparable<VirtualNode>
{
	final private File backingFile;
	final private Attributes attributes;
	final private VirtualDirectory parent;
	final VirtualFS virtualFS;

	public VirtualNode(File backingFile, VirtualDirectory parent, VirtualFS virtualFS)
	{
		this.backingFile = backingFile;
		this.attributes = VirtualUtils.parseFileName(backingFile.getName());
		this.parent = parent;
		this.virtualFS = virtualFS;
	}

	public boolean isDotFile()
	{
		return backingFile.getName().startsWith(".");
	}

	public boolean isTildeFile()
	{
		return backingFile.getName().endsWith("~");
	}

	public Attributes getAttributes()
	{
		return attributes;
	}

	public VirtualDirectory getParent()
	{
		return parent;
	}

	public File getBackingFile()
	{
		return backingFile;
	}

	public VirtualFS getVirtualFS()
	{
		return virtualFS;
	}
	
	public void setConflict()
	{
		String s1=getAttributes().createShort();
		int last=s1.lastIndexOf(".");
		if(last>=0)
		{
			String first=s1.substring(0,last);
			String end=s1.substring(last);
			attributes.setUrlName(first+"_"+getIndex(this)+end);
		}
		else
			attributes.setUrlName(s1+"_"+getIndex(this));
		
	}

	@Override
	public String toString()
	{
		return "File " + attributes.getVisibleName() + " / " + attributes.getSortNr() + " / "
				+ attributes.getName();
	}

	@Override
	public int compareTo(VirtualNode o)
	{
		long diff = attributes.getSortNr() - o.attributes.getSortNr();
		if (diff == 0)
			return attributes.getVisibleName().compareTo(o.attributes.getVisibleName());
		else
			return diff < 0 ? 1 : -1;
	}

	@Override
	public boolean equals(Object obj)
	{
		if (obj == this)
			return true;
		if (!(obj instanceof VirtualNode))
			return false;
		return compareTo((VirtualNode) obj) == 0;
	}

	public VirtualDirectory findDirectory(String path)
	{
		VirtualDirectory vdir = getDirectory(this);
		StringTokenizer st = new StringTokenizer(path, "/");
		while (st.hasMoreTokens())
		{
			String token = st.nextToken();
			VirtualDirectory vdir2 = vdir.getAccessibleChildDirectory(token);
			if (vdir2 == null)
				return null;
			else
				vdir = vdir2;
		}
		return vdir;
	}

	public Collection<? extends VirtualNode> find(String path)
	{
		Collection<VirtualNode> tmp = new ArrayList<VirtualNode>();
		StringTokenizer st = new StringTokenizer(path, "/");
		VirtualDirectory vdir = getDirectory(this);
		while (st.hasMoreTokens())
		{
			String token = st.nextToken();
			if (token.equals("*"))
			{
				tmp.addAll(vdir.getAccessibleDirectories());
				tmp.addAll(vdir.getAccessibleFiles());
				return tmp;
			}
			else
			{
				VirtualDirectory vdir2 = vdir.getAccessibleChildDirectory(token);
				if (vdir2 == null)
					return null;
				else
					vdir = vdir2;
			}
		}
		tmp.add(vdir);
		return tmp;
	}

	private VirtualDirectory getDirectory(VirtualNode node)
	{
		if (node instanceof VirtualDirectory)
			return (VirtualDirectory) node;
		else
		{
			VirtualFile vfile = (VirtualFile) node;
			return vfile.getParent();
		}
	}

	public String getRelativeVisibleFile() throws UnsupportedEncodingException
	{
		StringBuilder sb = new StringBuilder("/");
		sb.append(URLEncoder.encode(getAttributes().getVisibleName(), "UTF-8"));
		if (this instanceof VirtualDirectory)
			sb.append("/");
		VirtualDirectory vdir = getParent();
		if (vdir == null)
			return "/";
		while (vdir.getParent() != null)
		{
			sb.insert(0, URLEncoder.encode(vdir.getAttributes().getVisibleName(), "UTF-8"));
			sb.insert(0, "/");
			vdir = vdir.getParent();
		}
		return sb.toString();
	}
	
	public static String getIndex(VirtualNode virtualNode)
	{
		CRC32 crc = new CRC32();
		crc.update(virtualNode.getAttributes().getName().getBytes());
		long hash = Math.abs(crc.getValue() ^ virtualNode.getAttributes().getSortNr());
		return Long.toString(hash);
	}
	
}
